@@ -1,7 +1,7 @@
 /*
- * lftp and utils
+ * lftp - file transfer program
  *
- * Copyright (c) 1996-2010 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2013 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,12 +14,9 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/* $Id: CmdExec.cc,v 1.152 2011/03/25 14:35:41 lav Exp $ */
-
 #include <config.h>
 #include <stddef.h>
 #include <errno.h>
@@ -57,9 +54,16 @@ static ResType lftp_cmd_vars[] = {
    {"cmd:verify-path-cached",	 "no",	  ResMgr::BoolValidate,0},
    {"cmd:verify-host",		 "yes",	  ResMgr::BoolValidate,0},
    {"cmd:at-exit",		 "",	  0,0},
+   {"cmd:at-exit-bg",		 "",	  0,0},
+   {"cmd:at-exit-fg",		 "",	  0,0},
+   {"cmd:at-background",	 "",	  0,0},
+   {"cmd:at-terminate",		 "",	  0,0},
+   {"cmd:at-finish",		 "",	  0,0},
+   {"cmd:at-queue-finish",	 "",	  0,0},
    {"cmd:fail-exit",		 "no",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"cmd:verbose",		 "no",	  ResMgr::BoolValidate,ResMgr::NoClosure},
-   {"cmd:interactive",		 "no",	  ResMgr::BoolValidate,ResMgr::NoClosure},
+   {"cmd:interactive",		 "auto",  ResMgr::TriBoolValidate,ResMgr::NoClosure},
+   {"cmd:show-status",		 "yes",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"cmd:move-background",	 "yes",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"cmd:move-background-detach","yes",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"cmd:set-term-status",	 "no",	  ResMgr::BoolValidate,0},
@@ -67,6 +71,7 @@ static ResType lftp_cmd_vars[] = {
    {"cmd:trace",		 "no",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {"cmd:parallel",		 "1",	  ResMgr::UNumberValidate,0},
    {"cmd:queue-parallel",	 "1",	  ResMgr::UNumberValidate,0},
+   {"cmd:cls-exact-time",	 "yes",	  ResMgr::BoolValidate,ResMgr::NoClosure},
    {0}
 };
 static ResDecls lftp_cmd_vars_register(lftp_cmd_vars);
@@ -156,6 +161,7 @@ void  CmdExec::exec_parsed_command()
    case(COND_ANY):
       if(exit_code!=0 && ResMgr::QueryBool("cmd:fail-exit",0))
       {
+         failed_exit_code=exit_code;
 	 while(feeder)
 	    RemoveFeeder();
 	 cmd_buf.Empty();
@@ -194,6 +200,8 @@ restart:
 
    const struct cmd_rec *c;
    const char *cmd_name=args->getarg(0);
+   if(!cmd_name)
+      return;
    int part=find_cmd(cmd_name,&c);
    if(part<=0)
       eprintf(_("Unknown command `%s'.\n"),cmd_name);
@@ -251,6 +259,7 @@ void CmdExec::AddNewJob(Job *new_job)
    if(new_job->jobno<0)
       new_job->AllocJobno();
    new_job->SetParentFg(this,!background);
+   exit_code=0;
    AddWaiting(new_job);
    if(background) {
       Roll(new_job);
@@ -291,7 +300,7 @@ int CmdExec::Done()
 {
    Enter();
    bool done = (feeder==0 && Idle())
-      || (auto_terminate_in_bg && NumberOfJobs()==0 && !in_foreground_pgrp());
+      || (auto_terminate_in_bg && NumberOfChildrenJobs()==0 && !in_foreground_pgrp());
    Leave();
    return done;
 }
@@ -314,6 +323,7 @@ void CmdExec::RemoveFeeder()
       queue_feeder=0;
    delete replace_value(feeder,feeder->prev);
    Reconfig(0);
+   SetInteractive();
 }
 
 void CmdExec::ReuseSavedSession()
@@ -341,6 +351,17 @@ void CmdExec::ChangeSlot(const char *n)
       ChangeSession(s->Clone());
 }
 
+void CmdExec::AtFinish()
+{
+   if(queue_feeder && queue_feeder->JobCount())
+      return;
+   if(!fed_at_finish && NumAwaitedJobs()==0 && cmd_buf.Size()==0) {
+      FeedCmd(ResMgr::Query(queue_feeder?"cmd:at-queue-finish":"cmd:at-finish",0));
+      FeedCmd("\n");
+      fed_at_finish=true;
+   }
+}
+
 int CmdExec::Do()
 {
    int m=STALL;
@@ -402,11 +423,7 @@ int CmdExec::Do()
 		  session->Close();
 		  exit_code=0;
 		  builtin=BUILTIN_NONE;
-		  char *cmd=string_alloca(6+3+2*strlen(loc));
-		  strcpy(cmd,"open \"");
-		  unquote(cmd+strlen(cmd),loc);
-		  strcat(cmd,"\";");
-		  PrependCmd(cmd);
+		  PrependCmd(xstring::get_tmp("open ").append_quoted(loc));
 	       	  return MOVED;
 	       }
 	    }
@@ -524,7 +541,7 @@ int CmdExec::Do()
 	    return MOVED;
 	 }
       }
-      if(status_line && status_line->CanShowNow())
+      if(status_line && show_status && status_line->CanShowNow())
 	 ShowRunStatus(status_line);   // this is only for top level CmdExec.
       return m;
    }
@@ -568,9 +585,9 @@ int CmdExec::Do()
 	    return MOVED;
 	 }
       }
-      if(status_line && status_line->CanShowNow())
+      if(status_line && show_status && status_line->CanShowNow())
 	 ShowRunStatus(status_line);   // this is only for top level CmdExec.
-      if(m != STALL || interactive || !feeder || waiting_num >= max_waiting)
+      if(m != STALL || interactive || waiting_num >= max_waiting)
 	 return m;
    }
 
@@ -594,7 +611,7 @@ try_get_cmd:
 	       last_bg=-1;
 	 }
 
-	 if(status_line)
+	 if(status_line && show_status)
 	 {
 	    const char *def_title = FormatPrompt(ResMgr::Query("cmd:default-title",getenv("TERM")));
 	    status_line->DefaultTitle(def_title);
@@ -636,6 +653,9 @@ try_get_cmd:
 	 {
 	    auto_terminate_in_bg=false;
 	    FeedCmd(cmd);
+	    Roll();
+	    if(!Idle())
+	       fed_at_finish=false;
 	    return MOVED;
 	 }
 	 else
@@ -651,8 +671,7 @@ try_get_cmd:
       return m;
    }
 
-   parse_result
-      res = parse_one_cmd();
+   parse_result res = parse_one_cmd();
 
    if(alias_field<=0)
       free_used_aliases();
@@ -669,8 +688,10 @@ try_get_cmd:
 	 feeder->Bg();
       break;
    }
-   if(args==0 || args->count()==0)
+   if(args==0 || args->count()==0) {
+      AtFinish();
       return MOVED;  // empty command
+   }
 
    if(interactive)
       session->DontSleep(); // We don't want to get a delay just after user
@@ -706,58 +727,49 @@ void CmdExec::ShowRunStatus(const SMTaskRef<StatusLine>& s)
    }
 }
 
-void CmdExec::PrintStatus(int v,const char *prefix)
+xstring& CmdExec::FormatStatus(xstring& s,int v,const char *prefix)
 {
-   SessionJob::PrintStatus(v,prefix);
+   SessionJob::FormatStatus(s,v,prefix);
    if(builtin)
    {
-      xstring_ca s(args->Combine());
-      printf(_("\tExecuting builtin `%s' [%s]\n"),s.get(),session->CurrentStatus());
-      return;
+      xstring_ca ac(args->Combine());
+      return s.appendf(_("\tExecuting builtin `%s' [%s]\n"),ac.get(),session->CurrentStatus());
    }
    if(queue_feeder)
    {
       if(IsSuspended())
-	 printf("%s%s\n",prefix,_("Queue is stopped."));
+	 s.appendf("%s%s\n",prefix,_("Queue is stopped."));
       BuryDoneJobs();
       for(int i=0; i<waiting_num; i++)
       {
 	 if(i==0)
-	    printf("%s%s ",prefix,_("Now executing:"));
+	    s.appendf("%s%s ",prefix,_("Now executing:"));
 	 if(v==0)
-	    waiting[i]->ListOneJob(v);
+	    waiting[i]->FormatOneJob(s,v);
 	 else
-	    waiting[i]->PrintJobTitle();
+	    waiting[i]->FormatJobTitle(s);
 	 if(i+1<waiting_num)
-	    printf("%s\t-",prefix);
+	    s.appendf("%s\t-",prefix);
       }
-      queue_feeder->PrintStatus(v,prefix);
-      return;
+      return queue_feeder->FormatStatus(s,v,prefix);
    }
    if(waiting_num==1)
-   {
-      printf(_("\tWaiting for job [%d] to terminate\n"),waiting[0]->jobno);
-      return;
-   }
+      return s.appendf(_("\tWaiting for job [%d] to terminate\n"),waiting[0]->jobno);
    else if(waiting_num>1)
    {
-      printf(_("\tWaiting for termination of jobs: "));
+      s.appendf(_("\tWaiting for termination of jobs: "));
       for(int i=0; i<waiting_num; i++)
       {
-	 printf("[%d]",waiting[i]->jobno);
-	 printf("%c",i+1<waiting_num?' ':'\n');
+	 s.appendf("[%d]",waiting[i]->jobno);
+	 s.append(i+1<waiting_num?' ':'\n');
       }
-      return;
+      return s;
    }
    if(cmd_buf.Size()>0)
-   {
-      // xgettext:c-format
-      printf(_("\tRunning\n"));
-   }
+      s.append(_("\tRunning\n"));
    else if(feeder)
-   {
-      printf(_("\tWaiting for command\n"));
-   }
+      s.append(_("\tWaiting for command\n"));
+   return s;
 }
 
 void CmdExec::init(LocalDirectory *c)
@@ -769,6 +781,7 @@ void CmdExec::init(LocalDirectory *c)
    background=false;
 
    interactive=false;
+   show_status=true;
    top_level=false;
    auto_terminate_in_bg=false;
    feeder=0;
@@ -781,7 +794,9 @@ void CmdExec::init(LocalDirectory *c)
    condition=COND_ANY;
    prev_exit_code=0;
    exit_code=0;
+   failed_exit_code=0;
    last_bg=-1;
+   fed_at_finish=true;
 
    cwd=c;
    if(!cwd)
@@ -807,12 +822,12 @@ void CmdExec::init(LocalDirectory *c)
    Reconfig();
 }
 CmdExec::CmdExec(FileAccess *s,LocalDirectory *c)
-   : SessionJob(s?s:new DummyProto)
+   : SessionJob(s?s:new DummyProto), parent_exec(0)
 {
    init(c);
 }
-CmdExec::CmdExec(const CmdExec *parent)
-   : SessionJob(parent->session->Clone())
+CmdExec::CmdExec(CmdExec *parent)
+   : SessionJob(parent->session->Clone()), parent_exec(parent)
 {
    init(parent->cwd->Clone());
 }
@@ -885,11 +900,7 @@ const char *CmdExec::FormatPrompt(const char *scan)
 const char *CmdExec::MakePrompt()
 {
    if(partial_cmd)
-   {
-      static char partial_prompt[] = "> ";
-      return partial_prompt;
-   }
-
+      return "> ";
    return FormatPrompt(ResMgr::Query("cmd:prompt",getenv("TERM")));
 }
 
@@ -899,6 +910,7 @@ void CmdExec::beep_if_long()
    && now.UnixTime()>start_time+long_running
    && interactive && Idle() && isatty(1))
       write(1,"\007",1);
+   AtFinish();
 }
 
 void CmdExec::Reconfig(const char *name)
@@ -914,10 +926,11 @@ void CmdExec::Reconfig(const char *name)
    verify_path_cached=ResMgr::QueryBool("cmd:verify-path-cached",c);
    verify_host=ResMgr::QueryBool("cmd:verify-host",c);
    verbose=ResMgr::QueryBool("cmd:verbose",0);
-   // only allow explicit setting of cmd:interactive to change interactiveness.
-   if(top_level && name && !strcmp(name,"cmd:interactive"))
-      SetInteractive(ResMgr::QueryBool("cmd:interactive",0));
-   max_waiting=ResMgr::Query(queue_feeder?"cmd:queue-parallel":"cmd:parallel",c);
+   if(top_level || queue_feeder)
+      max_waiting=ResMgr::Query(queue_feeder?"cmd:queue-parallel":"cmd:parallel",c);
+   if(name && !strcmp(name,"cmd:interactive"))
+      SetInteractive();
+   show_status=ResMgr::QueryBool("cmd:show-status",0);
 }
 
 void CmdExec::pre_stdout()
@@ -941,6 +954,7 @@ void CmdExec::SetCmdFeeder(CmdFeeder *new_feeder)
    new_feeder->saved_buf.set(cmd_buf.Get());
    feeder=new_feeder;
    cmd_buf.Empty();
+   SetInteractive();
 }
 
 int CmdExec::AcceptSig(int sig)
@@ -1022,48 +1036,47 @@ void CmdExec::SetInteractive(bool i)
    }
    interactive=i;
 }
-
-int CmdExec::unquote(char *buf,const char *str)
+void CmdExec::SetInteractive()
 {
-   char *buf0=buf;
-   while(*str)
+   if(!top_level)
+      return;
+   bool def=feeder?feeder->IsInteractive():false;
+   SetInteractive(ResMgr::QueryTriBool("cmd:interactive",0,def));
+}
+
+xstring& xstring::append_quoted(const char *str,int len)
+{
+   if(!CmdExec::needs_quotation(str,len))
+      return append(str);
+
+   append('"');
+   while(len>0)
    {
       if(*str=='"' || *str=='\\')
-	 *buf++='\\';
-      *buf++=*str++;
+	 append('\\');
+      append(*str++);
+      len--;
    }
-   *buf=0;
-   return buf-buf0;
+   return append('"');
 }
 
-const char *CmdExec::unquote(const char *str)
+bool CmdExec::needs_quotation(const char *buf,int len)
 {
-   static xstring ret;
-   ret.get_space(strlen(str)*2);
-   ret.set_length(unquote(ret.get_non_const(),str));
-   return ret;
-}
-
-bool CmdExec::needs_quotation(const char *buf)
-{
-   while(*buf)
+   while(len>0)
    {
       if(*buf==' ' || *buf=='\t')
 	 return true;
       if(strchr("\"'\\&|>;",*buf))
 	 return true;
       buf++;
+      len--;
    }
    return false;
 }
 
 void CmdExec::FeedQuoted(const char *c)
 {
-   char *buf=(char*)alloca(strlen(c)*2+2+1);
-   buf[0]='"';
-   unquote(buf+1,c);
-   strcat(buf,"\"");
-   FeedCmd(buf);
+   FeedCmd(xstring::get_tmp("").append_quoted(c));
 }
 
 // implementation is here because it depends on CmdExec.
@@ -1075,10 +1088,7 @@ char *ArgV::CombineQuoted(int start) const
    for(;;)
    {
       const char *arg=String(start++);
-      if(CmdExec::needs_quotation(arg))
-	 res.vappend("\"",CmdExec::unquote(arg),"\"",NULL);
-      else
-	 res.append(arg);
+      res.append_quoted(arg);
       if(start>=Count())
 	 return(res.borrow());
       res.append(' ');
@@ -1107,6 +1117,27 @@ void CmdExec::AtExit()
    status_line=0;
 }
 
+void CmdExec::AtExitBg()
+{
+   FeedCmd(ResMgr::Query("cmd:at-exit-bg",0));
+   FeedCmd("\n");
+}
+void CmdExec::AtExitFg()
+{
+   FeedCmd(ResMgr::Query("cmd:at-exit-fg",0));
+   FeedCmd("\n");
+}
+void CmdExec::AtBackground()
+{
+   FeedCmd(ResMgr::Query("cmd:at-background",0));
+   FeedCmd("\n");
+}
+void CmdExec::AtTerminate()
+{
+   FeedCmd(ResMgr::Query("cmd:at-terminate",0));
+   FeedCmd("\n");
+}
+
 void CmdExec::EmptyCmds()
 {
    cmd_buf.Empty();
@@ -1225,10 +1256,14 @@ Job *CmdExec::default_cmd()
 }
 Job *CmdExec::builtin_local()
 {
+   if(args->count()<2) {
+      eprintf(_("Usage: %s cmd [args...]\n"),args->a0());
+      return 0;
+   }
    saved_session=session.borrow();
    session=FileAccess::New("file");
    if(!session) {
-      eprintf("%s: cannot create local session\n",args->a0());
+      eprintf(_("%s: cannot create local session\n"),args->a0());
       RevertToSavedSession();
       return 0;
    }
