@@ -695,7 +695,7 @@ bool FileCopy::CheckFileSizeAtEOF() const
    const long long get_pos=get->GetRealPos();
    const long long put_pos=put->GetRealPos();
    const long long pos=(get_pos>put_pos ? get_pos : put_pos);
-   if(pos>=range_limit)
+   if(pos<=0 || pos>=range_limit)
       return true;
 
    debug((0,"expected pos=%lld, actual pos=%lld\n",range_limit,pos));
@@ -910,14 +910,7 @@ int FileCopyPeerFA::Do()
 		  return m;
 	       if(res==FA::STORE_FAILED)
 	       {
-		  try_time=session->GetTryTime();
-		  retries=session->GetRetries();
-		  off_t pos=session->GetRealPos();
-		  if(upload_watermark<pos) {
-		     upload_watermark=pos;
-		     retries=-1;
-		  }
-		  Log::global->Format(10,"try_time=%ld, retries=%d\n",(long)try_time,retries);
+		  upload_state.Save(session);
 		  session->Close();
 		  if(can_seek && seek_pos>0)
 		     Seek(FILE_END);
@@ -1051,16 +1044,15 @@ void FileCopyPeerFA::OpenSession()
    session->Open(file,FAmode,seek_pos);
    session->SetFileURL(orig_url);
    session->SetLimit(range_limit);
-   if(mode==PUT)
-   {
-      if(try_time!=NO_DATE)
-	 session->SetTryTime(try_time);
-      if(retries>=0)
-	 session->SetRetries(retries+1);
+   if(mode==PUT) {
+      upload_state.Restore(session);
       if(e_size!=NO_SIZE && e_size!=NO_SIZE_YET)
 	 session->SetSize(e_size);
       if(date!=NO_DATE && date!=NO_DATE_YET)
 	 session->SetDate(date);
+   } else {
+      if(size!=NO_SIZE && size!=NO_SIZE_YET)
+	 session->SetSize(size);
    }
    session->RereadManual();
    if(ascii)
@@ -1134,7 +1126,7 @@ int FileCopyPeerFA::Get_LL(int len)
 	    if(FAmode==FA::QUOTE_CMD)
 	       FAmode=FA::RETRIEVE;
 
-	    xstring& loc=xstring::get_tmp(loc_c);
+	    xstring loc(loc_c);
 	    session->Close(); // loc_c is no longer valid.
 	    loc_c=0;
 
@@ -1184,8 +1176,7 @@ int FileCopyPeerFA::Get_LL(int len)
 	    size=NO_SIZE_YET;
 	    date=NO_DATE_YET;
 
-	    try_time=NO_DATE;
-	    retries=-1;
+	    upload_state.Clear();
 	    current->Timeout(0); // retry with new location.
 	    return 0;
 	 }
@@ -1195,6 +1186,7 @@ int FileCopyPeerFA::Get_LL(int len)
    }
    else if(res==0)
    {
+      debug((10,"copy-peer: EOF on %s\n",session->GetFileURL(session->GetFile())));
       eof=true;
       FileAccess::cache->Add(session,file,FAmode,FA::OK,this);
       SetSuggestedFileName(session->GetSuggestedFileName());
@@ -1231,14 +1223,7 @@ int FileCopyPeerFA::Put_LL(const char *buf,int len)
 	 return 0;
       if(res==FA::STORE_FAILED)
       {
-	 try_time=session->GetTryTime();
-	 retries=session->GetRetries();
-	 off_t pos=session->GetRealPos();
-	 if(upload_watermark<pos) {
-	    upload_watermark=pos;
-	    retries=-1;
-	 }
-	 Log::global->Format(10,"try_time=%ld, retries=%d\n",(long)try_time,retries);
+	 upload_state.Save(session);
 	 session->Close();
 	 if(can_seek && seek_pos>0)
 	    Seek(FILE_END);
@@ -1292,9 +1277,6 @@ void FileCopyPeerFA::Init()
 {
    get_delay=0;
    fxp=false;
-   try_time=NO_DATE;
-   retries=-1;
-   upload_watermark=0;
    redirections=0;
    can_seek=true;
    can_seek0=true;
@@ -1670,8 +1652,10 @@ int FileCopyPeerFDStream::Get_LL(int len)
    }
 #endif
 
-   if(res==0)
+   if(res==0) {
+      debug((10,"copy-peer: EOF on FD %d\n",fd));
       eof=true;
+   }
    return res;
 }
 
@@ -1902,7 +1886,8 @@ int FileCopyPeerMemory::Do()
 void FileVerificator::Init0()
 {
    done=false;
-   if(!ResMgr::QueryBool("xfer:verify",0))
+   if(!ResMgr::QueryBool("xfer:verify",0)
+   || ResMgr::Query("xfer:verify-command",0).is_empty())
       done=true;
 }
 void FileVerificator::InitVerify(const char *f)
