@@ -182,6 +182,8 @@ xstring& MirrorJob::FormatShortStatus(xstring& s)
 {
    if(bytes_to_transfer>0 && (!parent_mirror || parent_mirror->bytes_to_transfer!=bytes_to_transfer)) {
       long long curr_bytes_transferred=GetBytesCount();
+      if(parent_mirror)
+         curr_bytes_transferred+=bytes_transferred;
       s.appendf("%s/%s (%d%%)",
 	 xhuman(curr_bytes_transferred),xhuman(bytes_to_transfer),
 	 percent(curr_bytes_transferred,bytes_to_transfer));
@@ -209,6 +211,18 @@ void MirrorJob::TransferFinished(Job *j)
    AddBytesTransferred(bytes_count);
    stats.bytes+=bytes_count;
    stats.time +=j->GetTimeSpent();
+   if(j->ExitCode()==0 && verbose_report>=2) {
+      xstring finished;
+      const xstring& cmd=j->GetCmdLine();
+      if(cmd[0]=='\\')
+	 finished.append(cmd+1,cmd.length()-1);
+      else
+	 finished.append(cmd);
+      const xstring& rate=Speedometer::GetStrProper(j->GetTransferRate());
+      if(rate.length()>0)
+	 finished.append(" (").append(rate).append(')');
+      Report(_("Finished %s"),finished.get());
+   }
    JobFinished(j);
    if(transfer_count==0)
       root_mirror->transfer_time_elapsed += now-root_mirror->transfer_start_ts;
@@ -228,12 +242,18 @@ void  MirrorJob::HandleFile(FileInfo *file)
    struct stat st;
 
    // TODO: get rid of local hacks.
+
    // dir_name returns pointer to static data - need to dup it.
    const char *source_name=dir_file(source_dir,file->name);
    source_name=alloca_strdup(source_name);
    const char *target_name=dir_file(target_dir,file->name);
    target_name=alloca_strdup(target_name);
 
+   const char *source_name_rel=dir_file(source_relative_dir,file->name);
+   source_name_rel=alloca_strdup(source_name_rel);
+   const char *target_name_rel=dir_file(target_relative_dir,file->name);
+   target_name_rel=alloca_strdup(target_name_rel);
+
    FileInfo::type filetype=FileInfo::NORMAL;
    if(file->defined&file->TYPE)
       filetype=file->filetype;
@@ -291,8 +311,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    }
 	    else if(!to_rm_mismatched->FindByName(file->name))
 	    {
-	       Report(_("Removing old file `%s'"),
-			dir_file(target_relative_dir,file->name));
+	       Report(_("Removing old file `%s'"),target_name_rel);
 	       remove_target=true;
 	       stats.mod_files++;
 	    }
@@ -301,15 +320,13 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 }
 	 else if(flags&ONLY_EXISTING)
 	 {
-	    Report(_("Skipping file `%s' (only-existing)"),
-		     dir_file(source_relative_dir,file->name));
+	    Report(_("Skipping file `%s' (only-existing)"),source_name_rel);
 	    goto skip;
 	 }
 	 else
 	    stats.new_files++;
 
-	 Report(_("Transferring file `%s'"),
-		  dir_file(source_relative_dir,file->name));
+	 Report(_("Transferring file `%s'"),source_name_rel);
 
 	 if(script)
 	 {
@@ -360,7 +377,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 if(file->defined&file->SIZE)
 	    cp->SetSize(file->size);
 	 TransferStarted(cp);
-	 cp->cmdline.vset("\\transfer ",file->name.get(),NULL);
+	 cp->cmdline.vset("\\transfer `",source_name_rel,"'",NULL);
 
 	 set_state(WAITING_FOR_TRANSFER);
 	 break;
@@ -378,8 +395,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 {
 	    if(flags&ONLY_EXISTING)
 	    {
-	       Report(_("Skipping directory `%s' (only-existing)"),
-			dir_file(target_relative_dir,file->name));
+	       Report(_("Skipping directory `%s' (only-existing)"),target_name_rel);
 	       goto skip;
 	    }
 	 }
@@ -398,8 +414,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	       }
 	       else
 	       {
-		  Report(_("Removing old local file `%s'"),
-			   dir_file(target_relative_dir,file->name));
+		  Report(_("Removing old local file `%s'"),target_name_rel);
 		  if(remove(target_name)==-1)
 		  {
 		     eprintf("mirror: remove(%s): %s\n",target_name,strerror(errno));
@@ -415,7 +430,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    source_session->Clone(),target_session->Clone(),
 	    source_name,target_name);
 	 AddWaiting(mj);
-	 mj->cmdline.vset("\\mirror ",file->name.get(),NULL);
+	 mj->cmdline.vset("\\mirror `",source_name_rel,"'",NULL);
 
 	 // inherit flags and other things
 	 mj->SetFlags(flags,1);
@@ -423,8 +438,8 @@ void  MirrorJob::HandleFile(FileInfo *file)
 
 	 mj->SetExclude(exclude);
 
-	 mj->source_relative_dir.set(dir_file(source_relative_dir,file->name));
-	 mj->target_relative_dir.set(dir_file(target_relative_dir,file->name));
+	 mj->source_relative_dir.set(source_name_rel);
+	 mj->target_relative_dir.set(target_name_rel);
 
 	 mj->verbose_report=verbose_report;
 	 mj->newer_than=newer_than;
@@ -475,8 +490,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    FileInfo *old=target_set->FindByName(file->name);
 	    if(old && !to_rm_mismatched->FindByName(file->name))
 	    {
-	       Report(_("Removing old file `%s'"),
-			dir_file(target_relative_dir,file->name));
+	       Report(_("Removing old file `%s'"),target_name_rel);
 	       remove_target=true;
 	       stats.mod_symlinks++;
 	    }
@@ -507,8 +521,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    struct stat st;
 	    if(lstat(target_name,&st)!=-1)
 	    {
-	       Report(_("Removing old local file `%s'"),
-			dir_file(target_relative_dir,file->name));
+	       Report(_("Removing old local file `%s'"),target_name_rel);
 	       stats.mod_symlinks++;
 	       if(remove(target_name)==-1)
 	       {
@@ -520,14 +533,12 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    {
 	       if(flags&ONLY_EXISTING)
 	       {
-		  Report(_("Skipping symlink `%s' (only-existing)"),
-			   dir_file(target_relative_dir,file->name));
+		  Report(_("Skipping symlink `%s' (only-existing)"),target_name_rel);
 		  goto skip;
 	       }
 	       stats.new_symlinks++;
 	    }
-	    Report(_("Making symbolic link `%s' to `%s'"),
-		     dir_file(target_relative_dir,file->name),file->symlink.get());
+	    Report(_("Making symbolic link `%s' to `%s'"),target_name_rel,file->symlink.get());
 	    res=symlink(file->symlink,target_name);
 	    if(res==-1)
 	       eprintf("mirror: symlink(%s): %s\n",target_name,strerror(errno));
@@ -589,10 +600,16 @@ void  MirrorJob::InitSets(const FileSet *source,const FileSet *dest)
    if(!(flags&DELETE))
       to_transfer->SubtractAny(to_rm_mismatched);
 
-   to_transfer->SortByPatternList(ResMgr::Query("mirror:order",0));
-   to_transfer->CountBytes(&bytes_to_transfer);
-   if(parent_mirror)
-      parent_mirror->AddBytesToTransfer(bytes_to_transfer);
+   const char *sort_by=ResMgr::Query("mirror:sort-by",0);
+   bool desc=strstr(sort_by,"-desc");
+   if(!strncmp(sort_by,"name",4))
+      to_transfer->SortByPatternList(ResMgr::Query("mirror:order",0));
+   else if(!strncmp(sort_by,"date",4))
+      to_transfer->Sort(FileSet::BYDATE);
+   else if(!strncmp(sort_by,"size",4))
+      to_transfer->Sort(FileSet::BYSIZE,false,true);
+   if(desc)
+      to_transfer->ReverseSort();
 
    int dir_count=0;
    to_transfer->Count(&dir_count,NULL,NULL,NULL);
@@ -851,10 +868,16 @@ int   MirrorJob::Do()
       stats.dirs++;
 
       InitSets(source_set,target_set);
+
+      to_transfer->CountBytes(&bytes_to_transfer);
+      if(parent_mirror)
+	 parent_mirror->AddBytesToTransfer(bytes_to_transfer);
+
       to_rm->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
       to_rm->rewind();
       to_rm_mismatched->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
       to_rm_mismatched->rewind();
+
       set_state(TARGET_REMOVE_OLD_FIRST);
       goto TARGET_REMOVE_OLD_FIRST_label;
 
@@ -938,16 +961,16 @@ int   MirrorJob::Do()
 	       goto pre_TARGET_CHMOD;
 	    goto pre_WAITING_FOR_TRANSFER;
 	 }
+	 const char *target_name_rel=dir_file(target_relative_dir,file->name);
+	 target_name_rel=alloca_strdup(target_name_rel);
 	 if(!(flags&DELETE))
 	 {
 	    if(flags&REPORT_NOT_DELETED)
 	    {
-	       if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
-		  Report(_("Old directory `%s' is not removed"),
-			   dir_file(target_relative_dir,file->name));
+	       if(file->TypeIs(file->DIRECTORY))
+		  Report(_("Old directory `%s' is not removed"),target_name_rel);
 	       else
-		  Report(_("Old file `%s' is not removed"),
-			   dir_file(target_relative_dir,file->name));
+		  Report(_("Old file `%s' is not removed"),target_name_rel);
 	    }
 	    continue;
 	 }
@@ -984,16 +1007,10 @@ int   MirrorJob::Do()
 		  j->Recurse();
 	    }
 	 }
-	 if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
-	 {
-	    Report(_("Removing old directory `%s'"),
-		     dir_file(target_relative_dir,file->name));
-	 }
+	 if(file->TypeIs(file->DIRECTORY))
+	    Report(_("Removing old directory `%s'"),target_name_rel);
 	 else
-	 {
-	    Report(_("Removing old file `%s'"),
-		     dir_file(target_relative_dir,file->name));
-	 }
+	    Report(_("Removing old file `%s'"),target_name_rel);
       }
       break;
 
@@ -1111,6 +1128,7 @@ int   MirrorJob::Do()
    pre_DONE:
       set_state(DONE);
       m=MOVED;
+      bytes_transferred=0;
       if(!parent_mirror && (flags&LOOP) && stats.HaveSomethingDone(flags) && !stats.error_count)
       {
 	 PrintStatus(0,"");
