@@ -274,6 +274,11 @@ lftp_ssl_gnutls::lftp_ssl_gnutls(int fd1,handshake_mode_t m,const char *h)
    const char *auth=ResMgr::Query("ftp:ssl-auth", hostname);
    if(auth && !strncmp(auth, "SSL", 3))
       gnutls_priority_set_direct(session, "NORMAL:+SSL3.0:-TLS1.0:-TLS1.1:-TLS1.2",0);
+
+   if(h && ResMgr::QueryBool("ssl:use-sni",h)) {
+      if(gnutls_server_name_set(session, GNUTLS_NAME_DNS, h, xstrlen(h)) < 0)
+	 fprintf(stderr,"WARNING: failed to configure server name indication (SNI) TLS extension\n");
+   }
 }
 void lftp_ssl_gnutls::load_keys()
 {
@@ -317,6 +322,7 @@ lftp_ssl_gnutls::~lftp_ssl_gnutls()
    if(cred)
       gnutls_certificate_free_credentials(cred);
    gnutls_deinit(session);
+   session=0;
 }
 
 /* This function will try to verify the peer's certificate chain, and
@@ -354,7 +360,7 @@ void lftp_ssl_gnutls::verify_certificate_chain(const gnutls_datum_t *cert_chain,
       if(!gnutls_x509_crt_check_hostname(cert[0], hostname))
 	 set_cert_error(xstring::format("certificate common name doesn't match requested host name %s",quote(hostname)));
    } else {
-      Log::global->Format(0, "WARNING: Certificate verification: hostname checking disabled");
+      Log::global->Format(0, "WARNING: Certificate verification: hostname checking disabled\n");
    }
 
    for (i = 0; i < cert_chain_length; i++)
@@ -501,7 +507,7 @@ bool lftp_ssl_gnutls::check_fatal(int res)
    if((res==GNUTLS_E_UNEXPECTED_PACKET_LENGTH
        || res==GNUTLS_E_PUSH_ERROR || res==GNUTLS_E_PULL_ERROR
        || res==GNUTLS_E_DECRYPTION_FAILED)
-   && temporary_network_error(errno))
+   && (!errno || temporary_network_error(errno)))
       return false;
    return true;
 }
@@ -604,9 +610,11 @@ void lftp_ssl_gnutls::copy_sid(const lftp_ssl_gnutls *o)
 {
    size_t session_data_size;
    void *session_data;
-   gnutls_session_get_data(o->session,NULL,&session_data_size);
+   if(gnutls_session_get_data(o->session,NULL,&session_data_size)!=GNUTLS_E_SUCCESS)
+      return;
    session_data=xmalloc(session_data_size);
-   gnutls_session_get_data(o->session,session_data,&session_data_size);
+   if(gnutls_session_get_data(o->session,session_data,&session_data_size)!=GNUTLS_E_SUCCESS)
+      return;
    gnutls_session_set_data(session,session_data,session_data_size);
 }
 
@@ -825,6 +833,11 @@ lftp_ssl_openssl::lftp_ssl_openssl(int fd1,handshake_mode_t m,const char *h)
    ssl=SSL_new(instance->ssl_ctx);
    SSL_set_fd(ssl,fd);
    SSL_ctrl(ssl,SSL_CTRL_MODE,SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER,0);
+
+   if(h && ResMgr::QueryBool("ssl:use-sni",h)) {
+      if(!SSL_set_tlsext_host_name(ssl, h))
+	 fprintf(stderr,"WARNING: failed to configure server name indication (SNI) TLS extension\n");
+   }
 }
 void lftp_ssl_openssl::load_keys()
 {
@@ -861,6 +874,7 @@ void lftp_ssl_openssl::shutdown()
 lftp_ssl_openssl::~lftp_ssl_openssl()
 {
    SSL_free(ssl);
+   ssl=0;
 }
 
 static const char *verify_callback_host;
@@ -1242,7 +1256,7 @@ void lftp_ssl_openssl::check_certificate()
 
   bool check_hostname = ResMgr::QueryBool("ssl:check-hostname", hostname);
   if(!check_hostname) {
-    Log::global->Format(0, "WARNING: Certificate verification: hostname checking disabled");
+    Log::global->Format(0, "WARNING: Certificate verification: hostname checking disabled\n");
     return;
   }
 
