@@ -40,8 +40,13 @@
 # include <sys/time.h>
 #endif
 
+#if LIBIDN
+# include <idna.h>
+#endif
+
 CDECL_BEGIN
 #include "regex.h"
+#include "human.h"
 CDECL_END
 #include "misc.h"
 #include "ProcWait.h"
@@ -740,6 +745,8 @@ bool temporary_network_error(int err)
 {
    switch(err)
    {
+   case(EPIPE):
+   case(EIO):
    case(ETIMEDOUT):
 #ifdef ECONNRESET
    case(ECONNRESET):
@@ -989,3 +996,21 @@ const char *get_nodename()
       return u.nodename;
    return "NODE";
 }
+
+const char *xhuman(long long n)
+{
+   char *buf=xstring::tmp_buf(LONGEST_HUMAN_READABLE + 1);
+   return human_readable(n, buf, human_autoscale|human_SI, 1, 1);
+}
+
+const char *xidna_to_ascii(const char *name)
+{
+#if LIBIDN
+   if(!name)
+      return 0;
+   char *name_ace_tmp=0;
+   if(idna_to_ascii_lz(name,&name_ace_tmp,0)==IDNA_SUCCESS)
+      return xstring::get_tmp().set_allocated(name_ace_tmp);
+#endif//LIBIDN
+   return name;
+}
